package fr.itsonus.bousoleplussbackend.exception.handlers;

import fr.itsonus.bousoleplussbackend.domain.auth.exception.AlreadyExistingUserException;
import fr.itsonus.bousoleplussbackend.exception.InvalidPasswordException;
import fr.itsonus.bousoleplussbackend.exception.NoCurrentUserException;
import fr.itsonus.bousoleplussbackend.exception.TokenRefreshException;
import fr.itsonus.bousoleplussbackend.payload.response.ApiError;
import fr.itsonus.bousoleplussbackend.exception.UnknownEmailException;
import jakarta.validation.ConstraintViolationException;
import org.springframework.dao.DataIntegrityViolationException;
import org.springframework.data.rest.webmvc.ResourceNotFoundException;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.HttpStatusCode;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.core.AuthenticationException;
import org.springframework.web.bind.MethodArgumentNotValidException;
import org.springframework.web.bind.annotation.ControllerAdvice;
import org.springframework.web.bind.annotation.ExceptionHandler;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.client.RestClientResponseException;
import org.springframework.web.context.request.WebRequest;
import org.springframework.web.servlet.mvc.method.annotation.ResponseEntityExceptionHandler;

import java.util.ArrayList;

@ControllerAdvice
public class ExceptionsHandler extends ResponseEntityExceptionHandler {

    @ExceptionHandler({ConstraintViolationException.class})
    public ResponseEntity<ApiError> handleValidationExceptions(ConstraintViolationException ex) {
        var errors = new ArrayList<ApiError.FieldError>();
        ex.getConstraintViolations().forEach(error -> errors.add(new ApiError.FieldError(error.getMessage(), error.getPropertyPath().toString())));
        return new ApiError("Saisie invalide", errors).toResponse(HttpStatus.BAD_REQUEST);
    }

    @ExceptionHandler({AlreadyExistingUserException.class})
    public ResponseEntity<ApiError> handleConflictException(RuntimeException ex) {
        return new ApiError(ex.getMessage()).toResponse(HttpStatus.CONFLICT);
    }

    @ExceptionHandler({DataIntegrityViolationException.class})
    public ResponseEntity<ApiError> handleIntegrityViolationExceptions(DataIntegrityViolationException ex) {
        return new ApiError("Saisie invalide: " + ex.getMessage()).toResponse(HttpStatus.BAD_REQUEST);
    }

    @ExceptionHandler({InvalidPasswordException.class, UnknownEmailException.class})
    public ResponseEntity<ApiError> handleBadRequestException(RuntimeException ex) {
        return new ApiError(ex.getMessage()).toResponse(HttpStatus.BAD_REQUEST);
    }

    @ExceptionHandler({ResourceNotFoundException.class})
    public ResponseEntity<ApiError> handleResourceNotFoundException(ResourceNotFoundException ex) {
        return new ApiError(ex.getMessage()).toResponse(HttpStatus.NOT_FOUND);
    }

    @ExceptionHandler({AuthenticationException.class})
    public ResponseEntity<ApiError> handleAuthenticationException(Exception ex) {
        return new ApiError("Authentication failed: " + ex.getMessage()).toResponse(HttpStatus.UNAUTHORIZED);
    }

    @ExceptionHandler({AccessDeniedException.class, TokenRefreshException.class})
    public ResponseEntity<ApiError> handleAccessDeniedException(Exception ex) {
        return new ApiError("Authentication needed: " + ex.getMessage()).toResponse(HttpStatus.FORBIDDEN);
    }

    @ExceptionHandler({BadCredentialsException.class})
    public ResponseEntity<ApiError> handleBadCredentialsException(Exception ex) {
        return new ApiError("Vos identifiants sont incorrects.").toResponse(HttpStatus.UNAUTHORIZED);
    }

    @ExceptionHandler({NoCurrentUserException.class})
    public ResponseEntity<ApiError> handleInternalErrorException(RuntimeException ex) {
        logger.error("Internal error", ex);
        return new ApiError(ex.getMessage()).toResponse(HttpStatus.INTERNAL_SERVER_ERROR);
    }

    @ExceptionHandler({RestClientResponseException.class})
    @ResponseBody
    public ResponseEntity<ApiError> handleHttpClientErrorException(RestClientResponseException ex) {
        return new ApiError(ex.getStatusText()).toResponse(ex.getStatusCode());
    }

    @Override
    protected ResponseEntity<Object> handleMethodArgumentNotValid(MethodArgumentNotValidException ex,
                                                                  HttpHeaders headers, HttpStatusCode status, WebRequest request) {
        var errors = new ArrayList<ApiError.FieldError>();
        ex.getBindingResult().getFieldErrors()
                .forEach(error -> errors.add(new ApiError.FieldError(error.getDefaultMessage(), error.getField())));
        ex.getBindingResult().getGlobalErrors()
                .forEach(error -> errors.add(new ApiError.FieldError(error.getDefaultMessage(), error.getObjectName())));

        var apiError = new ApiError("Saisie invalide", errors);
        return ResponseEntity.status(status).body(apiError);
    }

    @ExceptionHandler({Throwable.class})
    @ResponseBody
    public ResponseEntity<ApiError> handleGlobally(Throwable ex) {
        if (logger.isDebugEnabled()) {
            logger.warn(ex.getMessage(), ex);
        } else {
            logger.warn(ex.getMessage());
        }
        return new ApiError(ex.getMessage()).toResponse(HttpStatus.INTERNAL_SERVER_ERROR);
    }

    @Override
    protected ResponseEntity<Object> handleExceptionInternal(Exception ex, Object body, HttpHeaders headers,
                                                             HttpStatusCode statusCode, WebRequest request) {
        logger.debug("An exception occurred", ex);
        return super.handleExceptionInternal(ex, new ApiError(ex.getMessage()), headers, statusCode, request);
    }
}
